sub vcl_recv {
  if (req.restarts == 0 && req.http.X-Authcache-Allow) {
    // Retrieve the authcache-key from /authcache-varnish-get-key before each
    // request. Upon vcl_deliver the authcacke-key is copied over to the
    // X-Authcache-Key request header and the request is restarted.
    set req.http.X-Original-URL = req.url;
    set req.url = req.http.X-Authcache-Get-Key-URL;
    set req.http.X-Authcache-Get-Key = 1;

    unset req.http.X-Authcache-Allow;
  }

  if (req.http.X-Authcache-Key || req.http.X-Authcache-Get-Key) {
    // Hide cookies from default vcl_recv.
    if (req.http.Cookie) {
      set req.http.X-Cookie-Backup = req.http.Cookie;
      unset req.http.Cookie;
    }
  }

  if (req.esi_level > 0 && req.http.X-Authcache-Do-ESI) {
    // Add authcache header for ESI requests
    set req.http.X-Authcache = 1;
  }
  else {
    unset req.http.X-Authcache;
  }
}

sub vcl_hash {
  // Restore cookies if they were hidden above.
  if (req.http.X-Cookie-Backup) {
    set req.http.Cookie = req.http.X-Cookie-Backup;
    unset req.http.X-Cookie-Backup;
  }

  // Separate cache bins for js/non-js clients
  if (req.http.Cookie ~ "(^|;)\s*has_js=1\s*($|;)") {
    hash_data("has_js=1");
  }
}

sub vcl_fetch {
  // Turn on ESI processing when required by backend
  if (beresp.http.X-Authcache-Do-ESI && req.http.X-Authcache-Do-ESI) {
    set beresp.do_esi = true;
  }

  // Do not cache a response when X-Authcache-Key is empty or missing.
  if (!req.http.X-Authcache-Key && !req.http.X-Authcache-Get-Key) {
    set beresp.ttl = 120 s;
    return (hit_for_pass);
  }

  // Do not cache when backend specifies Cache-Control: no-cache.
  if (beresp.http.Cache-Control ~ "(private|no-cache|no-store)") {
    set beresp.ttl = 120 s;
    return (hit_for_pass);
  }
}

sub vcl_deliver {
  // Process response from authcache-key callback
  if (req.http.X-Authcache-Get-Key) {

    // Restore the original URL.
    set req.url = req.http.X-Original-URL;
    unset req.http.X-Original-URL;

    // Copy over the X-Authcache-Key header if set
    if (resp.http.X-Authcache-Key) {
      set req.http.X-Authcache-Key = resp.http.X-Authcache-Key;
    }

    // Key-retrieval phase is over, remove the flag
    unset req.http.X-Authcache-Get-Key;

    return (restart);
  }

  // When sending the response back to the browser:
  // 1. Add "Cookie" to Vary header
  if (resp.http.Vary !~ "Cookie") {
    set resp.http.Vary = resp.http.Vary + ", Cookie";
  }
  // 2. Remove X-Authcache-Key from Vary header
  if (resp.http.Vary ~ "X-Authcache-Key") {
    set resp.http.Vary = regsub(resp.http.Vary, "(^|,\s*)X-Authcache-Key", "");
  }
  // 3. Remove a "," prefix, if present.
  set resp.http.Vary = regsub(resp.http.Vary, "^,\s*", "");

  // Remove variables placed on backend response.
  unset resp.http.X-Authcache-Do-ESI;
}
